/* Nes/Snes/Genesis/SMS/Atari to USB
 * Copyright (C) 2006-2007 Raphal Assnat
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * The author may be contacted at raph@raphnet.net
 */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <string.h>
#include "gamepad.h"
#include "leds.h"
#include "db9.h"

#define REPORT_SIZE		3
#define GAMEPAD_BYTES	3

/* for chaning IO easily */


//#define SAMPLE()		(PINB & 0x3f)
inline unsigned char SAMPLE()
{
	unsigned char c;
	unsigned char b;
	unsigned char res;

	c = PINC;
	b = PINB;

	/* The rest of the code was developped on a different PCB which
	 * maps the DB9 pins differently. This code is to 'emulate'
	 * this different mapping so I dont have to change auto-detection
	 * and report descriptor building stuff. */

	res = (c & 0x20) >> 5; // Up/Up/Z
	res |= (c & 0x10) >> 3; // Down/Down/Y
	res |= (c & 0x08) >> 1; // Left/0/X
	res |= (c & 0x04) << 1; // Right/0
	res |= (c & 0x02) << 3; // BtnB/BtnA
	res |= (b & 0x20); // BtnC/BtnStart

	return res;
}


#define SET_SELECT()	PORTC |= 1;
#define CLR_SELECT()	PORTC &= 0xfe;

/*********** prototypes *************/
static void db9Init(void);
static void db9Update(void);
static char db9Changed(void);
static void db9BuildReport(unsigned char *reportBuffer);



#define CTL_ID_ATARI	0x00 // up/dn/lf/rt/btn_b
#define CTL_ID_SMS		0x01 // up/dn/lf/rt/btn_b/btn_c
#define CTL_ID_GENESIS3	0x02 // up/dn/lf/rt/btn_a/btn_b/btn_c/start
#define CTL_ID_GENESIS6 0x03 // all bits

static unsigned char cur_id = CTL_ID_GENESIS3;

// the most recent bytes we fetched from the controller
static unsigned char last_read_controller_bytes[REPORT_SIZE];
// the most recently reported bytes
static unsigned char last_reported_controller_bytes[REPORT_SIZE];

static void readController(unsigned char bits[3])
{
	unsigned char a,b,c;

	// total delays: 160 microseconds.

	SET_SELECT();
	_delay_us(20.0);
	a = SAMPLE();

	if (cur_id == CTL_ID_ATARI ||
		cur_id == CTL_ID_SMS) {
		
		bits[0] = a;
		bits[1] = 0xff;
		bits[2] = 0xff;

		return;
	}

	CLR_SELECT();
	_delay_us(20.0);		
	b = SAMPLE();

	SET_SELECT();
	_delay_us(20.0);		
	CLR_SELECT();
	_delay_us(20.0);		
	SET_SELECT();
	_delay_us(20.0);		
	CLR_SELECT();
	_delay_us(20.0);			
	SET_SELECT();
	
	_delay_us(20.0);			

	c = SAMPLE();

	CLR_SELECT();
	_delay_us(20.0);			
	SET_SELECT();

	bits[0] = a;
	bits[1] = b;
	bits[2] = c;
}

static void db9Init(void)
{
	unsigned char sreg;
	unsigned char bits[3];
	sreg = SREG;
	cli();

	DDRB |= 0x02; // Bit 1 out
	PORTB &= ~0x02; // 0

	/* 
	 * --- nes_snes_pcb pinout ---
	 * PC5: Up/Up/Z
	 * PC4: Down/Down/Y
	 * PC3: Left/0/X
	 * PC2: Right/0
	 *
	 * PC1: BtnB/BtnA
	 * PC0: Select
	 *
	 * PB5: BtnC/BtnStart
	 * 
	 * --- db9 specific PCB pinout ---
	 * PB0: Up/Up/Z
	 * PB1: Down/Down/Y
	 * PB2: Left/0/X
	 * PB3: Right/0
	 *
	 * PB4: BtnB/BtnA
	 * PB5: BtnC/BtnStart
	 *
	 * PC0: Select output
	 *
	 */

	DDRB &= ~0x20; // PB5, BtnC/BtnStart
	PORTB |= 0x20;

	DDRC |= 1;	// SELECT
	PORTC |= 1;

	DDRC &= ~0x3E; // direction and buttons
	PORTC |= 0x3E;

	readController(bits);

	// when select was low, Left and Right should follow if this is
	// a genesis controller. 
	if (bits[1] & 0xc) {
		// change controller id. this will keep SELECT high for Atari 2600 joysticks which
		// may want to draw some current from this pin.
		cur_id = CTL_ID_SMS;
	}
	
	
	db9Update();

	SREG = sreg;
}



static void db9Update(void)
{
	unsigned char data[3];
	int x=128,y=128;
	
	/* 0: Up//Z
	 * 1: Down//Y
	 * 2: Left//X
	 * 3: Right
	 * 4: Btn B/A
	 * 5: Btn C/Start
	 */
	readController(data);

	/***** 6 Button auto-detection *****/
	if (cur_id == CTL_ID_GENESIS3) {
		unsigned char tmp, tmp2, tmp3;
		unsigned char data2[3];
		unsigned char data3[3];
		
		// if the X,Y or Z button is active
		// while the corresponding direction is not,
		// we might be dealing with a 6 button controller...
		tmp = data[0] & (data[2] ^ 7) & 7;
		if (tmp)
		{
			// ... but the user could have pressed a direction button
			// just at the right moment. So re-read the controller on 
			// or two  more times. (for a total of 3)
			//
			// Double polling will only be done on rare occasions with 3 button 
			// controllers. (when the user presses a direction at the moment we 
			// read XYZ buttons). It adds a 1.5ms delay.
			//
			// For triple polling to occur on 3 button controllers (3ms delay), the user would have
			// to press and release a direction button before 1.5ms passes, and then press
			// it again right in the middle of the 2nd poll... And to cause and erroneous 6 button
			// controller detection, he would have to repeat this exploit 1.5ms later for the 3rd poll.. 
			// I think this is very unlikely, unless the user is a signal generator ;)
			//
			// 6 Button controllers dont response to the SELECT line afer the 4pulse
			// read for a little more than 1ms. It is therefore necessary to inserta a delay
			// between additional polls:
			//
			// 1000us not enough
			// 1500us works
			// 2000us is probably seems safe. 
			_delay_ms(1.5); 
			readController(data2);

			tmp2 = data2[0] & (data2[2] ^ 7) & tmp;
			if (tmp2) {
				_delay_ms(1.5);
				readController(data3);
				tmp3 = data3[0] & (data3[2] ^ 7) & tmp;
				
				if (tmp3) {
					cur_id = CTL_ID_GENESIS6;
				}
			}
		}
	}


	/* Buttons are active high. Invert values. */
	data[0] = data[0] ^ 0xff;
	data[1] = data[1] ^ 0xff;
	data[2] = data[2] ^ 0xff;

	if (data[0] & 1) { y = 0; } // up
	if (data[0] & 2) { y = 255; } //down
	if (data[0] & 4) { x = 0; }  // left
	if (data[0] & 8) { x = 255; } // right

	last_read_controller_bytes[0]=x;
	last_read_controller_bytes[1]=y;
 
 	last_read_controller_bytes[2]=0;

	if (data[1]&0x10) { last_read_controller_bytes[2] |= 0x01; } // A
	if (data[0]&0x10) { last_read_controller_bytes[2] |= 0x02; } // B
	if (data[0]&0x20) { last_read_controller_bytes[2] |= 0x04; } // C
	if (data[1]&0x20) { last_read_controller_bytes[2] |= 0x08; } // Start
	if (cur_id == CTL_ID_GENESIS6) {
		if (data[2]&0x04) { last_read_controller_bytes[2] |= 0x10; } // X
		if (data[2]&0x02) { last_read_controller_bytes[2] |= 0x20; } // Y
		if (data[2]&0x01) { last_read_controller_bytes[2] |= 0x40; } // Z
	}
}	

static char db9Changed(void)
{
	static int first = 1;
	if (first) { first = 0;  return 1; }
	
	return memcmp(last_read_controller_bytes, 
					last_reported_controller_bytes, GAMEPAD_BYTES);
}

static void db9BuildReport(unsigned char *reportBuffer)
{
	if (reportBuffer != NULL)
	{
		memcpy(reportBuffer, last_read_controller_bytes, GAMEPAD_BYTES);
	}
	memcpy(last_reported_controller_bytes, 
			last_read_controller_bytes, 
			GAMEPAD_BYTES);	
}

#include "snes_descriptor.c"

Gamepad db9Gamepad = {
	report_size: 		REPORT_SIZE,
	reportDescriptorSize:	sizeof(snes_usbHidReportDescriptor),
	init: 			db9Init,
	update: 		db9Update,
	changed:		db9Changed,
	buildReport:		db9BuildReport
};

Gamepad *db9GetGamepad(void)
{
	db9Gamepad.reportDescriptor = (void*)snes_usbHidReportDescriptor;

	return &db9Gamepad;
}

